import requests  # type: ignore

from pybuoy.mixins.parser import ParserMixin  # type: ignore


class ApiBase(ParserMixin):
    def make_request(self, url: str):
        response = requests.get(url)
        if response.status_code == 200:
            return response.text
        elif response.status_code == 404:
            print(f"Data not available for url: {url}")
            return None
        else:
            raise ValueError(f"Error code {response.status_code} for url: {url}")



"""Mapping key to unit for Observation."""
# https://www.ndbc.noaa.gov/measdes.shtml
from enum import Enum
from typing import TypedDict


class BaseKey(Enum):
    ...


class MeteorologicalKey(BaseKey):
    WDIR = "WDIR"
    WSPD = "WSPD"
    GST = "GST"
    WVHT = "WVHT"
    DPD = "DPD"
    APD = "APD"
    MWD = "MWD"
    PRES = "PRES"
    ATMP = "ATMP"
    WTMP = "WTMP"
    DEWP = "DEWP"
    VIS = "VIS"
    PTDY = "PTDY"
    TIDE = "TIDE"


class WaveSummaryKey(BaseKey):
    WVHT = "WVHT"
    SwH = "SwH"
    SwP = "SwP"
    WWH = "WWH"
    WWP = "WWP"
    SwD = "SwD"
    WWD = "WWD"
    STEEPNESS = "STEEPNESS"
    APD = "APD"
    MWD = "MWD"


class SummaryWaveSteepnessValues(Enum):
    VERY_STEEP = "VERY_STEEP"
    SWELL = "SWELL"
    AVERAGE = "AVERAGE"
    NA = "N/A"


class MeasurementsAndUnits(TypedDict):
    label: str
    unit: str


DATE_RECORDED = {
    "#YY": {"label": "Year", "unit": "Y"},
    "MM": {"label": "Month", "unit": "m"},
    "DD": {"label": "Day", "unit": "d"},
    "hh": {"label": "Hour", "unit": "H"},
    "mm": {"label": "Minute", "unit": "M"},
}

# Realtime (dataset="txt")
METEOROLOGICAL: dict[MeteorologicalKey, MeasurementsAndUnits] = {
    MeteorologicalKey.WDIR: {"label": "Wind Direction", "unit": "degrees"},
    MeteorologicalKey.WSPD: {"label": "Wind Speed", "unit": "m/s"},
    MeteorologicalKey.GST: {"label": "Wind Gust", "unit": "m/s"},
    MeteorologicalKey.WVHT: {"label": "Wave Height", "unit": "m"},
    MeteorologicalKey.DPD: {"label": "Dominate Wave Period", "unit": "seconds"},
    MeteorologicalKey.APD: {"label": "Average Wave Period", "unit": "seconds"},
    MeteorologicalKey.MWD: {"label": "Wave Direction", "unit": "degrees"},
    MeteorologicalKey.PRES: {"label": "Sea Level Pressure", "unit": "hPa"},
    MeteorologicalKey.ATMP: {"label": "Air Temperature", "unit": "celcius"},
    MeteorologicalKey.WTMP: {"label": "Water Temperature", "unit": "celcius"},
    MeteorologicalKey.DEWP: {"label": "Dewpoint Temperature", "unit": "celcius"},
    MeteorologicalKey.VIS: {"label": "Visibility", "unit": "nmi"},  # nautical miles
    MeteorologicalKey.PTDY: {
        "label": "Pressure Tendency",
        "unit": "hPa",
    },  # direction +/- amount of pressure change over 3hr period
    MeteorologicalKey.TIDE: {"label": "Tide", "unit": "ft"},
}
# Realtime (dateset="spec")
WAVE_SUMMARY: dict[WaveSummaryKey, MeasurementsAndUnits] = {
    WaveSummaryKey.WVHT: {"label": "Significant Wave Height", "unit": "meters"},
    WaveSummaryKey.SwH: {"label": "Swell Height", "unit": "meters"},
    WaveSummaryKey.SwP: {"label": "Swell Period", "unit": "seconds"},
    WaveSummaryKey.WWH: {"label": "Wind Wave Height", "unit": "meters"},
    WaveSummaryKey.WWP: {"label": "Wind Wave Period", "unit": "seconds"},
    WaveSummaryKey.SwD: {
        "label": "Swell Direction",
        "unit": "string",  # cardinal directions (e.g., SE)
    },
    WaveSummaryKey.WWD: {
        "label": "Wind Wave Direction",
        "unit": "string",  # cardinal directions (e.g., SE)
    },
    WaveSummaryKey.STEEPNESS: {
        "label": "Steepness",
        "unit": "string",  # ? refer to enum class?
    },
    WaveSummaryKey.APD: {"label": "Average Wave Period", "unit": "seconds"},
    WaveSummaryKey.MWD: {
        "label": "Dominant Wave Direction",
        "unit": "degrees",
    },  # direction of waves at dominant period (DPD)
}


# TODO: update below for dynamic mapping

WAVE_UNITS = {
    "sea_surface_wave_significant_height": "meters",
    "sea_surface_wave_peak_period": "seconds",
    "sea_surface_wave_mean_period": "seconds",
    "sea_surface_swell_wave_significant_height": "meters",
    "sea_surface_swell_wave_period": "seconds",
    "sea_surface_wind_wave_significant_height": "meters",
    "sea_surface_wind_wave_period": "seconds",
    "sea_water_temperature": "celcius",
    "sea_surface_wave_to_direction": "degrees",
    "sea_surface_swell_wave_to_direction": "degrees",
    "sea_surface_wind_wave_to_direction": "degrees",
    "number_of_frequencies": "count",
    "center_frequencies": "Hz",
    "bandwidths": "Hz",
    "spectral_energy": "m**2/Hz",
    "mean_wave_direction": "degrees",
    "principal_wave_direction": "degrees",
    "polar_coordinate_r1": "1",
    "polar_coordinate_r2": "1",
    "calculation_method": None,
    "sampling_rate": "Hz",
}

CURRENT_UNITS = {
    "bin": "count",
    "depth": "meters",
    "direction_of_sea_water_velocity": "degrees",
    "sea_water_speed": "c/s",
    "upward_sea_water_velocity": "c/s",
    "error_velocity": "c/s",
    "platform_orientation": "degrees",
    "platform_pitch_angle": "degrees",
    "platform_roll_angle": "degrees",
    "sea_water_temperature": "celcius",
    "pct_good_3_beam": "percent",
    "pct_good_4_beam": "percent",
    "pct_rejected": "percent",
    "pct_bad": "percent",
    "echo_intensity_beam1": "count",
    "echo_intensity_beam2": "count",
    "echo_intensity_beam3": "count",
    "echo_intensity_beam4": "count",
    "correlation_magnitude_beam1": "count",
    "correlation_magnitude_beam2": "count",
    "correlation_magnitude_beam3": "count",
    "correlation_magnitude_beam4": "count",
    "quality_flags": None,
}

WIND_UNITS = {
    "wind_from_direction": "degrees",
    "wind_speed": "m/s",
    "wind_speed_of_gust": "m/s",
    "upward_air_velocity": "m/s",
}

# REALTIME_WAVE_UNITS = {}

# SPECTRAL_UNITS = {}



from pybuoy.api.base import ApiBase
from pybuoy.const import API_PATH, Endpoints


class Realtime(ApiBase):
    # TODO: map phrases like "meterological" to dataset (i.e., "txt")
    def get(self, station_id: str, dataset="txt"):
        """Get realtime data from the NDBC.

        There are six different data sources:
            - data_spec     Raw Spectral Wave Data
            - ocean         Oceanographic Data
            - spec          Spectral Wave Summary Data
            - supl          Supplemental Measurements Data
            - swdir         Spectral Wave Data (alpha1)
            - swdir2        Spectral Wave Data (alpha2)
            - swr1          Spectral Wave Data (r1)
            - swr2          Spectral Wave Data (r2)
            - txt           Standard Meteorological Data

        Args:
            buoy_id (str): id of buoy
            dataset (str): one of the aforementioned data sources
        """
        if not station_id:
            raise ValueError("station_id must be provided")

        dataset_options = {
            "data_spec",
            "ocean",
            "spec",
            "supl",
            "swdir",
            "swdir2",
            "swr1",
            "swr2",
            "txt",
        }

        if dataset not in dataset_options:
            raise ValueError(f"Dataset must be one of {', '.join(dataset_options)}")

        url = f"{API_PATH[Endpoints.REALTIME.value]}/{station_id}.{dataset}"
        return self.parse(self.make_request(url), dataset)



from pybuoy.observation.observation import MeteorologicalObservation  # noqa: F401
from pybuoy.observation.observation import WaveSummaryObservation  # noqa: F401
from pybuoy.observation.observation_datum import (  # noqa: F401
    ObservationFloatDatum,
    ObservationStringDatum,
)
from pybuoy.observation.observations import Observations  # noqa: F401



from typing import Iterator

from pybuoy.observation import MeteorologicalObservation, WaveSummaryObservation


class Observations:
    """Observations class encapsulates `Observation`s by datetime."""

    def __init__(
        self,
        observations: list[MeteorologicalObservation | WaveSummaryObservation] = None,
    ):
        # TODO: improve error handling
        self._data = observations if observations is not None else []
        self._size = len(self._data)

    @property
    def reports(self) -> list[MeteorologicalObservation | WaveSummaryObservation]:
        return self._data

    @reports.setter
    def reports(
        self, observation: MeteorologicalObservation | WaveSummaryObservation
    ) -> None:
        self._data.append(observation)
        self._size = len(self._data)

    def __getitem__(
        self, key: int
    ) -> MeteorologicalObservation | WaveSummaryObservation:
        return self._data[key]

    def __iter__(self) -> Iterator[MeteorologicalObservation | WaveSummaryObservation]:
        return iter(self._data)

    def __repr__(self) -> str:
        if len(self._data) < 1000:
            return "{}({})".format(
                self.__class__.__name__,
                ", ".join("{}={!r}".format(k, v) for k, v in self.__dict__.items()),
            )
        # Get the first and last three items
        items_to_display = self._data[:3] + self._data[-3:]
        # Find the which item has the longest repr
        max_length_repr = max(items_to_display, key=lambda x: len(repr(x)))
        # Get the length of the item with the longest repr
        padding = len(repr(max_length_repr))
        # Create a list of the reprs of each item and apply the padding
        values = [repr(item).rjust(padding) for item in items_to_display]
        # Insert the '...' inbetween the 3rd and 4th item
        values.insert(3, "...")
        # Convert the list to a string joined by commas
        array_as_string = ", ".join(values)
        return f"Observations([{array_as_string}])"



from pybuoy.api.base import ApiBase
from pybuoy.const import API_PATH, Endpoints


class Stations(ApiBase):
    def get_active(self):
        default_dataset = "xml"
        url = f"{API_PATH[Endpoints.ACTIVE_STATIONS.value]}.{default_dataset}"
        return self._clean_activestation_data(self.make_request(url))



# TODO: incorporate to improve performance
class MeteorologicalMeta(type):
    def __new__(cls, name, bases, attrs):
        attrs["__slots__"] = (
            "wind_direction",
            "wind_speed",
            "wind_gust",
            "wave_height",
            "dominate_wave_period",
            "average_wave_period",
            "wave_direction",
            "sea_level_pressure",
            "air_temperature",
            "water_temperature",
            "dewpoint_temperature",
            "visibility",
            "pressure_tendency",
            "tide",
            "_datetime",
        )
        return super(MeteorologicalMeta, cls).__new__(cls, name, bases, attrs)


class WaveMeta(type):
    def __new__(cls, name, bases, attrs):
        attrs["__slots__"] = (
            "sea_surface_wave_significant_height",
            "sea_surface_wave_peak_period",
            "sea_surface_wave_mean_period",
            "sea_surface_swell_wave_significant_height",
            "sea_surface_swell_wave_period",
            "sea_surface_wind_wave_significant_height",
            "sea_surface_wind_wave_period",
            "sea_water_temperature",
            "sea_surface_wave_to_direction",
            "sea_surface_swell_wave_to_direction",
            "sea_surface_wind_wave_to_direction",
            "number_of_frequencies",
            "center_frequencies",
            "bandwidths",
            "spectral_energy",
            "mean_wave_direction",
            "principal_wave_direction",
            "polar_coordinate_r1",
            "polar_coordinate_r2",
            "calculation_method",
            "sampling_rate",
            "_datetime",
        )
        return super(WaveMeta, cls).__new__(cls, name, bases, attrs)



"""Provide the ParserMixin class."""
from datetime import datetime as dt
from xml.etree.cElementTree import Element, fromstring

from pybuoy.const import RealtimeDatasets
from pybuoy.observation import (
    MeteorologicalObservation,
    Observations,
    WaveSummaryObservation,
)
from pybuoy.unit_mappings import MeteorologicalKey, WaveSummaryKey


class XmlToDict(dict):
    def __init__(self, data: Element):
        if data.items():
            self.update(dict(data.items()))
        for element in data:
            if element:
                # treat like dict - assumes that if the first two tags
                # in a series are different, then all are different.
                if len(element) == 1 or element[0].tag != element[1].tag:
                    tmp_dict: dict[str, XmlToDict] | XmlToDict = XmlToDict(element)
                # treat like list - we assume that if the first two tags
                # in a series are the same, then the rest are the same.
                else:
                    tmp_dict = {element[0].tag: XmlToDict(element)}
                # if the tag has attributes, add those to the dict
                if isinstance(element, XmlToDict) and element.items():
                    tmp_dict.update(dict(element.items()))
                self.update({element.tag: tmp_dict})
            # assumes an attribute in a tag without any text.
            elif element.items():
                self.update({element.tag: dict(element.items())})
            else:
                self.update({element.tag: element.text})


class ParserMixin:
    """Interface for Buoy classes and its composites."""

    def parse(self, data: str, dataset: str = RealtimeDatasets.TXT.value):
        return (
            data
            if dataset != RealtimeDatasets.TXT.value
            and dataset != RealtimeDatasets.SPEC.value
            else self.__clean_realtime_data(data=data, dataset=dataset)
        )

    def _clean_activestation_data(self, data: str):
        xml_tree = fromstring(data)
        return [XmlToDict(el) for el in xml_tree.findall("station")]

    def __clean_realtime_data(self, data: str, dataset: str) -> Observations:
        # TODO: Error handling when request not successful
        if data is None:
            return None
        rows = data.strip().split("\n")

        headers = [" ".join(row.split()).split(" ") for row in rows[0:2]]
        header_offset = 5  # end of datetime columns
        headers_without_dates = headers[0][header_offset:]

        # TODO: consider csv module
        realtime_data = Observations()
        for row in rows[2:]:
            record_array = " ".join(row.split()).split(" ")
            date_recorded = dt(
                int(record_array[0]),
                int(record_array[1]),
                int(record_array[2]),
                int(record_array[3]),
                int(record_array[4]),
            )

            observation: MeteorologicalObservation | WaveSummaryObservation
            values: dict[MeteorologicalKey, str] | dict[WaveSummaryKey, str]
            if dataset == RealtimeDatasets.TXT.value:
                values = {
                    MeteorologicalKey[headers_without_dates[idx]]: value
                    for idx, value in enumerate(record_array[header_offset:])
                }
                observation = MeteorologicalObservation(
                    values,
                    date_recorded,
                )
            elif dataset == RealtimeDatasets.SPEC.value:
                values = {
                    WaveSummaryKey[headers_without_dates[idx]]: value
                    for idx, value in enumerate(record_array[header_offset:])
                }
                observation = WaveSummaryObservation(
                    values,
                    date_recorded,
                )
            # TODO: refactor to use setter method
            realtime_data.reports.append(
                observation
            ) if observation is not None else None
        return realtime_data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybuoy', 'pybuoy.api', 'pybuoy.mixins', 'pybuoy.observation']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'pybuoy',
    'version': '0.4.0',
    'description': 'Python wrapper for NDBC data.',
    'long_description': "![PyPI - Version](https://img.shields.io/pypi/v/pybuoy?color=blue)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pybuoy)\n![PyPI - Monthly Downloads](https://img.shields.io/pypi/dm/pybuoy)\n\n# pybuoy\n\n`pybuoy` is a server-side Python package that was built to facilitate rapid discovery of new data from [NDBC](https://www.ndbc.noaa.gov/) with only a single dependency!\n\n## Installation\n\n`pybuoy` is supported on Python 3.10+ and can be installed with either `pip` or a package manager like [poetry](https://python-poetry.org):\n\n- **with pip**: `pip install pybuoy`\n  - recommended to install any third party library in python's virtualenv. ([ref](https://packaging.python.org/en/latest/guides/installing-using-pip-and-virtual-environments))\n- **with poetry**: `poetry add pybuoy`\n  - automatically creates and manages your virtualenvs. ([ref](https://realpython.com/dependency-management-python-poetry))\n\n## Quickstart\n\n```python\n# Demo in python/ipython shell\n# Don't forget to install pybuoy first\n\n>>> from pybuoy import Buoy\n\n>>> buoy = Buoy()\n\n>>> buoy\n<pybuoy.buoy.Buoy object at 0x10481fc10>\n```\n\n## Examples\n\n- [Get all active stations](./docs/examples/get_activestations.py).\n\n- [Get realtime meteorological data](./docs/examples/get_realtime_data.py) for buoy by station_id.\n",
    'author': 'Kyle J. Burda',
    'author_email': 'kylejbdev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from pybuoy.const import NO_VALUE
from pybuoy.unit_mappings import MeasurementsAndUnits, SummaryWaveSteepnessValues


class ObservationFloatDatum(float):
    """Encapsulates `Buoy` datum from `Observation` as float."""

    def __init__(self, value: str, label_unit: MeasurementsAndUnits):
        """Initialize ObservationDatum record with relevant metadata.

        Args:
            value (str): value of observation. (Default: "MM")
            label_unit (MeasurementsAndUnits): label and unit for value.
        """
        self.value = None if value == NO_VALUE else float(value)
        self._label = label_unit["label"]
        self._unit = label_unit["unit"]

    def __new__(cls, value, *_):
        return (
            float.__new__(cls, value)
            if value != NO_VALUE
            else float.__new__(cls, "nan")
        )

    @property
    def label(self):
        """Return type of observation."""
        return self._label

    @property
    def unit(self):
        """Return this observation's unit."""
        return self._unit

    def __repr__(self):
        return f"ObservationDatum({self.label}: {self.__float__()} {self.unit})"

    def __str__(self):
        return f"{self.__float__()} {self.unit}"


class ObservationStringDatum(str):
    """Encapsulates `Buoy` datum from `Observation` as string."""

    def __init__(self, value: str, label_unit: MeasurementsAndUnits):
        """Initialize ObservationDatum record with relevant metadata.

        Args:
            value (str): value of observation. (Default: "MM")
            label_unit (MeasurementsAndUnits): label and unit for value.
        """
        self.value = (
            None
            if value == NO_VALUE or value == SummaryWaveSteepnessValues.NA.value
            else value
        )
        self._label = label_unit["label"]
        self._unit = label_unit["unit"]

    def __new__(cls, value, *_):
        return str.__new__(cls, value)

    @property
    def label(self):
        """Return type of observation."""
        return self._label

    @property
    def unit(self):
        """Return this observation's unit."""
        return self._unit

    def __repr__(self):
        return f"ObservationDatum({self.label}: {self.__str__()})"



"""pybuoy constants."""
from enum import Enum

from pybuoy.endpoints import API_PATH as API_PATH  # noqa: F401


class Endpoints(Enum):
    ACTIVE_STATIONS = "active_stations"
    REALTIME = "realtime2"


class RealtimeDatasets(Enum):
    SPEC = "spec"
    TXT = "txt"


NO_VALUE = "MM"

__version__ = "0.4.0"
